#!/usr/bin/perl
use strict;
use Getopt::Long;

my $do_rc    = 1; ## turn on looking at reverse complements
GetOptions ("rc=i" => \$do_rc);

my $input_fa = $ARGV[0]; # fasta file
my $k        = $ARGV[1]; # the size of kmer

## read in the fasta file
my $count = 0;
open(IN, $input_fa) or die "could not open $input_fa: $_";
while(my $line=<IN>){
	chomp($line);
	if($line =~ /^>/){
		my $name = "dummy";
		($name) = split(/\s+/,$line);
		$name =~ s/>//g;
		my $seq = "";
		while(my $l2=<IN>){
			if($l2 !~ /^>/){
				chomp($l2);
				$seq .= $l2;
			}else{
				seek(IN, -length($l2), 1);
				last;
			}
		}
    		my $seq2 = $seq;
		$seq2 =~ s/N//g;
		my $len = length($seq2);
    		my $buffer1 = PrintKDistrToBuf($seq, $k);
    		print "$count\t$name\_$count\t$len$buffer1\n";
		$count++;
	}else{
		next;
	}
}
close(IN);    

sub PrintKDistrToBuf {
    my $seq = shift;
    my $k = shift;

    ## convert the A/C/G/T sequence to 0/1/2/3 sequence
    my @sarr = split("",$seq);
    my @sarrbin = ();
    for (my $i=0; $i<=$#sarr; $i++) {
	my $ch = $sarr[$i];
	if ($ch !~ /[AaCcGgTt]/) { push @sarrbin, 4; next; }
	if ($ch =~ /[Aa]/) { push @sarrbin, 0; next; }
	if ($ch =~ /[Cc]/) { push @sarrbin, 1; next; }
	if ($ch =~ /[Gg]/) { push @sarrbin, 2; next; }
	if ($ch =~ /[Tt]/) { push @sarrbin, 3; next; }
    }

    ## scan the sarrbin sequence in windows of length k
    my $kword = 0;
    my $flag  = 0; ## will be set to i if the ith position from right is a non 0/1/2/3

    my %counts;
    ## set up the first k-lengh word
    for (my $i=0; $i<$k; $i++) {
	## check if the next char is valid or not
	if ($sarrbin[$i] > 3) {
	    $flag = 1;
	}
	else{
	    if ($flag > 0) { $flag = ($flag+1)%($k+1); } ## increment, modulo k
	}
	## append to kword
	$kword = $kword*4 + ($sarrbin[$i]%4); ## append 0 for 4
    }
    if ($flag == 0) {
	$counts{$kword} ++;
    }

    ## set up its reverse complement (rc)
    my $kwordcopy = $kword;
    my $kwordrc = 0;
    for (my $i=0; $i<$k; $i++) {
	my $ch = $kwordcopy %4;
	if ($ch <= 3) { $ch = 3-$ch; }
	$kwordrc = $kwordrc*4 + $ch;
	$kwordcopy = int($kwordcopy/4);
    }
    if ($flag == 0) {
	if ($do_rc == 1) {
		$counts{$kwordrc} ++;
	}
    }
    #warn Base4($kword)."\t".Base4($kwordrc)."\t$flag\n";
    #getc();

    ## now start scanning
    my $max = 4**$k;
    my $mostsig = 4**($k-1);
    my $len = length($seq);
    for (my $i=$k; $i<($len-$k+1); $i++) {
	my $ch = $sarrbin[$i];
	## check if the next char is valid or not

	if ($flag > 0) { $flag = ($flag+1)%($k+1); } ## increment, modulo k
	if ($ch > 3) {
	    $flag = 1;
	}

	## update kwords
	$ch = $ch % 4; ## 0 for 4
	$kword = ($kword*4)%$max + $ch;
	$kwordrc = (3-$ch)*$mostsig + int($kwordrc/4);
	
	## update counts
	if ($flag == 0) {
	    $counts{$kword} ++;
	    if ($do_rc == 1) {
	    	$counts{$kwordrc}++;
	    }
	}
	#warn Base4($kword)."\t".Base4($kwordrc)."\t$flag\n";
	#getc();
    }

    ## now print buffer
    my $buf = "";
    for (my $i=0; $i<$max; $i++) {
	if (defined($counts{$i})) {
	    $buf = $buf . "\t". $counts{$i};
	}
	else {
	    $buf = $buf . "\t0";
	}
    }
    return $buf;
}

sub Base4 {
    my $num = shift;
    my @val = ();
    while ($num > 0) {
	my $ch = $num%4;
	push @val, $ch;
	$num = int($num/4);
    }
    my $v = join("",reverse(@val)); 
    return $v;

}
	
